# -*- coding: UTF-8 -*-

from o2dbErrors import *

def X(x):
    """Usage: cross([list1,list2]), cross([list1,list2,list3]), ..."""
    if x == []: return [[]]
    return [[z]+y for y in X(x[1:]) for z in x[0]]

class SearchAttribute:
    """
    e.g. searching a person in a special age -
    construct SA1 = SearchAttribute("CHAR", "Firstname", equal="John")
    construct a second one SA2 = SearchAttribute("CHAR", "Surname", equal="Steed")
    construct a third one SA3 = SearchAttribute("DATE", "DateOfBirth", greaterOrEqual="1980-01-01", less="2001-01-01")
    call the method of the SearchInterface with a _list_ of the three SearchAttributes
    """
    def __init__(self, connBroker, type, name, **values):
        self.CB = connBroker
        self.type = type
        self.name = name        
        self.searchTableList = []
        self.equal = self.greater = self.greaterOrEqual = None
        self.less = self.lessOrEqual = self.like = self.substring = None

        # doppelte anfhrungsstriche erlaube ich nicht - werden durch einfache ersetzt
        for k in values.keys():
            #print k, ':', attrValue[k]
            if k == 'equal':
                self.equal = values[k].replace('"', "'")
                self.searchTableList = [self.CB.selectTable(type, values[k])]
            elif k == 'greater':
                self.greater = values[k].replace('"', "'")
                self.searchTableList = [self.CB.selectTable(type, values[k])]
            elif k == 'greaterOrEqual':
                self.greaterOrEqual = values[k].replace('"', "'")
                self.searchTableList = [self.CB.selectTable(type, values[k])]    
            elif k == 'less':
                self.less = values[k].replace('"', "'")
                self.searchTableList = [self.CB.selectTable(type, values[k])]
            elif k == 'lessOrEqual':
                self.lessOrEqual = values[k].replace('"', "'")
                self.searchTableList = [self.CB.selectTable(type, values[k])]
            elif k == 'like':
                self.like = values[k].replace('"', "'")
                self.searchTableList = self.CB.selectTable(type, values[k], 1)
            elif k == 'substring':
                self.substring = values[k].replace('"', "'")
                self.searchTableList = self.CB.selectTable(type, values[k], 1)      
      
            else: raise ErrSearchParameter("error search parameter", "received unknown parameter - '" + k + "'")

	       
    def getSearchTableList(self):
        return self.searchTableList
    def getType(self):
        return self.type
    def getName(self):
        return self.name
    def getValues(self):
        if self.equal:
            return "=", self.equal
        if self.like:
            return " LIKE ", self.like + "%"
        elif self.greaterOrEqual and self.lessOrEqual:
            return ">=", self.greaterOrEqual, "<=", self.lessOrEqual
        elif self.greater and self.lessOrEqual:
            return ">", self.greater, "<=", self.lessOrEqual
        elif self.greaterOrEqual and self.less:
            return ">=", self.greaterOrEqual, "<", self.less      
        elif self.greater and self.less:
            return ">", self.greater, "<", self.less
        elif self.greaterOrEqual:
            return ">=", self.greaterOrEqual
        elif self.lessOrEqual:
            return "<=", self.lessOrEqual
        elif self.greater:
            return ">", self.greater
        elif self.less:
            return "<", self.less
        elif self.substring:
            return self.substring
       
class SearchInterface:        
    def __init__(self, connBroker):
        self.CB = connBroker
        self.DBtype = connBroker.getDBtype()
        self.conn = None
        
    def __del__(self):
        """return connection if still in hand"""
        self._returnConnection()
        
    def _getCursor(self):
        if self.conn:
            return self.conn.cursor()
        else:
            self.conn = self.CB.getConnection(id(self))
            return self.conn.cursor()

    def _returnConnection(self):
        if self.conn:
            self.conn = None
            self.CB.objReturnsConnection(id(self))

    def searchObjects(self, searchAttrs):
        if self.DBtype == "mySQL":
            """little test for all the character saving C hackers"""
            LOL = []
            for i in range(len(searchAttrs)): LOL.append(searchAttrs[i].getSearchTableList())
            #STL = X(LOL)
            STL = reduce( lambda X1,X2: [ one+[two] for one in X1 for two in X2], LOL, [[]] )
            #STL = reduce( lambda X1,X2: [ one+(two,) for one in X1 for two in X2], LOL, [()] )
            res = []
            c = self._getCursor()
            for i in range(len(STL)):
                r = self._searchObjects_mySQL(c, searchAttrs, STL[i])
                for j in range(len(r)): res.append(r[j])
            return res
            c.close()
            
        elif self.DBtype == "postgreSQL":
            return self._searchObjects_postgreSQL(searchAttrs)
        elif self.DBtype == "what_else_DB":
            return self._searchObjects_whatElseDB(searchAttrs)
        else:
            raise ErrSearchDBtype("error searching DB-type", "no SQL statements for '" + self.DBtype + "' available!")

    def _searchObjects_mySQL(self, cursor, searchAttrs, searchTableList):
        """
        select * from tblCHAR10 a1, tblCHAR10 a2, tblCHAR10 a3
        where (a1.objID = a2.objID and a2.objID = a3.objID)
        and (a1.attrName='Zuname'         and a1.attrValue='BERGER'
        and  a2.attrName='objectcategory' and a2.attrValue='Kunde'
        and  a3.attrName='Vorname'        and a3.attrValue='Josef')
        """
        l = len(searchAttrs)
        strSQL = 'SELECT a0.objID \nFROM '
        
        SA = searchAttrs[0]
        strSQL += searchTableList[0] + ' a0' 
        for i in range(1, l):
             strSQL += ', ' + searchTableList[i] + ' a' + str(i) 
                
        join = ''
        if l == 2:
            join = '(a0.objID=a1.objID)\nAND '
        elif l > 2:
            join = '(a0.objID = a1.objID'
            for i in range(2, l):
                join += ' AND a' + str(i-1) + '.objID = a' + str(i) + '.objID'
            join += ')\nAND '
         
        SA = searchAttrs[0]
        strSQL += '\nWHERE ' + join + '(a0.attrName="' + SA.getName() + '" AND a0.attrValue' + \
                  SA.getValues()[0] + '"' + SA.getValues()[1] + '"'
        try:
            strSQL += ' AND a0.attrValue' + SA.getValues()[2] + '"' + SA.getValues()[3] + '"'
        except: pass
        
        for i in range(1, l):
            SA = searchAttrs[i]
            strSQL += '\nAND a' + str(i) + '.attrName="' + SA.getName() + '" AND a' + str(i) + '.attrValue' + \
                      SA.getValues()[0] + '"' + SA.getValues()[1] + '"'
            try:
                strSQL += ' AND a' + str(i) + '.attrValue' + SA.getValues()[2] + '"' + SA.getValues()[3] + '"'
            except: pass
           
        strSQL += ')'

        ## print strSQL

        cursor.execute(strSQL)
        return cursor.fetchall()
    
    def _searchObjects_postgreSQL(self, searchAttrs):
        ## translate the SQL string for postgreSQL
        pass


    def searchFULLTEXT(self, searchValue):
        """search FULLTEXT in the fulltextfield of all objects"""

        # doppelte anfhrungsstriche erlaube ich nicht - werden durch einfache ersetzt
        sVal = searchValue.replace('"', "'")

        if self.DBtype == "mySQL":              
            return self._searchFULLTEXT_mySQL(sVal)                    
        elif self.DBtype == "postgreSQL":
            return self._searchFULLTEXT_postgreSQL(sVal)
        elif self.DBtype == "what_else_DB":
            return self._searchObjects_whatElseDB(sVal)
        else:
            raise ErrSearchDBtype("error searching DB-type", "no SQL statements for '" + self.DBtype + "' available!")

    def _searchFULLTEXT_mySQL(self, searchVal):
        """select * from tblfulltextindex where match (attrValue) against ('java')"""
        cur = self._getCursor()

        strSQL = 'SELECT * FROM tblfulltextindex WHERE MATCH (attrValue) AGAINST ("'
        strSQL += searchVal + '")'

        cur.execute(strSQL)
        result = cur.fetchall()
        
        cur.close()
        return result

    def _searchFULLTEXT_postgreSQL(self, searchVal):
        ## translate the SQL string for postgreSQL
        pass


    def searchSUBSTRING(self, searchAttrs):
        if self.DBtype == "mySQL":
            """little test for all the character saving C hackers"""
            LOL = []
            for i in range(len(searchAttrs)): LOL.append(searchAttrs[i].getSearchTableList())
            STL = X(LOL)
            res = []
            c = self._getCursor()
            for i in range(len(STL)):
                r = self._searchSUBSTRING_mySQL(c, searchAttrs, STL[i])
                for j in range(len(r)): res.append(r[j])
            return res
            c.close()
            
        elif self.DBtype == "postgreSQL":
            return self._searchSUBSTRING_postgreSQL(searchAttrs)
        elif self.DBtype == "what_else_DB":
            return self._searchSUBSTRING_whatElseDB(searchAttrs)
        else:
            raise ErrSearchDBtype("error searching DB-type", "no SQL statements for '" + self.DBtype + "' available!")

    def _searchSUBSTRING_mySQL(self, cursor, searchAttrs, searchTableList):
        """
        SELECT * FROM tblCHAR10 a1, tblCHAR25 a2, tblCHAR25 a3
        WHERE (a1.objID = a2.objID and a2.objID = a3.objID)
        AND (a1.attrName='objectcategory'  AND a1.attrValue='Customer'
        AND (a2.attrName='Surname' AND LOCATE('substring', a2.attrValue) <> 0
        AND  a3.attrName='Address' AND LOCATE('different substring', a3.attrValue) <> 0)
        """
        l = len(searchAttrs)
        strSQL = 'SELECT a0.objID \nFROM '
        
        SA = searchAttrs[0]
        strSQL += searchTableList[0] + ' a0' 
        for i in range(1, l):
             strSQL += ', ' + searchTableList[i] + ' a' + str(i) 
                
        join = ''
        if l == 2:
            join = '(a0.objID=a1.objID)\nAND '
        elif l > 2:
            join = '(a0.objID = a1.objID'
            for i in range(2, l):
                join += ' AND a' + str(i-1) + '.objID = a' + str(i) + '.objID'
            join += ')\nAND '
         
        SA = searchAttrs[0]
        strSQL += '\nWHERE ' + join + '(a0.attrName="' + SA.getName() + '" AND '

        if SA.getName() == "objectcategory":
            strSQL += 'a0.attrValue' + SA.getValues()[0] + '"' + SA.getValues()[1] + '"'
        else:
            strSQL += 'LOCATE(UPPER("' + SA.getValues() + '"), UPPER(a0.attrValue)) <> 0'
        
        for i in range(1, l):
            SA = searchAttrs[i]
            strSQL += '\nAND a' + str(i) + '.attrName="' + SA.getName() + '" AND '

            if SA.getName() == "objectcategory":
                strSQL += 'a' + str(i) + '.attrValue' + SA.getValues()[0] + '"' + SA.getValues()[1] + '"'
            else:
                strSQL += 'LOCATE(UPPER("' + SA.getValues() + '"), UPPER(a' + str(i) + '.attrValue)) <> 0'
           
        strSQL += ')'

        #print strSQL
        
        cursor.execute(strSQL)
        return cursor.fetchall()
